/* FILE:        BFDVPsimpleGLUT.cpp
 * DATE:        5/8/2014
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2014, BitFlow, Inc.
 * DESCRIPTION: A simplified interface for displaying a single grayscale image using FreeGLUT.
 */

#define NOMINMAX

#include "BFDVPsimpleGLUT.h"
#include "SimpleTimers.hpp"

// Standard includes.
#include <limits>
#include <cmath>

#include <atomic>
#include <mutex>

#include <string>

// OpenGL includes (for display).
#include <GL\glew.h>    
#ifdef _WIN32
#include <GL\wglew.h>
#endif

// GLUT includes.
#include <GL\freeglut.h>

// CUDA includes.
#include <cuda_runtime_api.h>
#include <cuda_gl_interop.h>

// The OpenGL display variables and function.
struct GLBufferParams
{
    GLuint                  pbo;
    GLuint                  tex;
    cudaGraphicsResource_t  pbo_resource;

    GLint                   width;
    GLint                   height;
    GLint                   bytedepth;
    size_t                  size;

    GLBufferParams (void)
        : pbo           (0)
        , tex           (0)
        , pbo_resource  (0)
        , width         (0)
        , height        (0)
        , bytedepth     (0)
        , size          (0)
    { }

    void clean (void)
    {
        if (pbo)
        {
            cudaGraphicsUnregisterResource(pbo_resource);
            glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
            glDeleteBuffers(1, &pbo);
        }

        if (tex)
            glDeleteTextures(1, &tex);
        
        pbo = 0;
        tex = 0;
        pbo_resource = 0;

        width = 0;
        height = 0;
        bytedepth = 0;
        size = 0;
    }

    void make (const int w, const int h, const int b)
    {
        clean();

        width = w;
        height = h;
        bytedepth = b;
        size = width * height * bytedepth;

        // allocate new buffers
        glEnable(GL_TEXTURE_2D);
        glGenTextures(1, &tex);
        glBindTexture(GL_TEXTURE_2D, tex);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        if (1 == bytedepth)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE8, width, height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, NULL);
        else if (2 == bytedepth)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE16, width, height, 0, GL_LUMINANCE, GL_UNSIGNED_SHORT, NULL);
        else if (3 == bytedepth)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);

        glGenBuffers(1, &pbo);
        glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
        glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB, size, NULL, GL_DYNAMIC_COPY);

        //While a PBO is registered to CUDA, it can't be used
        //as the destination for OpenGL drawing calls.
        //But in our particular case OpenGL is only used
        //to display the content of the PBO, specified by CUDA kernels,
        //so we need to register/unregister it only once.
        cudaGraphicsGLRegisterBuffer(&pbo_resource, pbo, cudaGraphicsMapFlagsWriteDiscard);
    }
};

// BFDVPsimpleGLUT global parameters.
GLBufferParams              dispBuffer;
std::atomic<CUdeviceptr>    cudaBufferAtm (0);

    // Externally accessible variables.
std::string                 overlayText;
RateAverager<>              bufferDisplayRate;
std::mutex                  externallyAccessibleVariableMutex;

BFDVPsimpleGLUTkeyCallback  keyCallback = nullptr;

GLfloat dispWidth, dispHeight;
GLfloat zoomScalar = 1.f;
GLint xOffset = 0, yOffset = 0;

bool smoothTransitionsEnabled = true;

// Callback to time display updates.
void gfsdTimerEventFunc (int value)
{
    // Continue updating the display at about 60Hz.
    glutPostRedisplay();
    glutTimerFunc(1000/60, gfsdTimerEventFunc, 0);
}

// Callback to update the viewport.
void gfsdReshapeFunc (int width, int height)
{
    dispWidth = GLfloat(width);
    dispHeight = GLfloat(height);
    glViewport(0, 0, width, height);
}

// Smoothly transition the value of a.
float transitionToValue (float &a, const float d, const float by)
{
    const float aInit = a;
    const float absolute = abs(a - d);
    const float maximum = absolute * by;

    if (!smoothTransitionsEnabled)
        a = d;
    else if (a > d)
        a -= std::min(maximum, absolute);
    else if (d > a)
        a += std::min(maximum, absolute);

    return a - aInit;
}

void renderText (std::string const& text, const int lineWidth, const int x, const int y, const float r, const float g, const float b, const float a = 1.f)
{
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(0.0, dispWidth, 0.0, dispHeight);

    glDisable(GL_TEXTURE_2D);
    
    glTranslatef(float(x), float(y), 0);
    glScalef(0.2f, 0.2f, 0.2f);
    
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(r, g, b, a);
    glLineWidth((GLfloat)lineWidth);
    
    glutStrokeString( GLUT_STROKE_MONO_ROMAN, (unsigned char*)text.c_str() );
    
    glDisable(GL_BLEND);
    
    glEnable(GL_TEXTURE_2D);
        
    glPopMatrix();
}

// Callback to display a buffer.
GLfloat displayScaling = 1.f, displayXOffset = 0.f, displayYOffset = 0.f;
void gfsdDisplayFunc (void)
{
    //// Setup the display area.
    
    // Reset the coordinate system.
    glLoadIdentity();

    // Clear the display to a light blue color.
    glClearColor(0.8f, 1.f, 1.f, 0.f);
    glClear(GL_COLOR_BUFFER_BIT);

    // Retrieve the buffer to copy for display, if any.
    const CUdeviceptr cudaBuffer = cudaBufferAtm.load();
    if (!cudaBuffer)
    {
        // Swap the front and back buffers.
        glutSwapBuffers();
        return;
    }

    //// Copy the CUDA buffer.

    // Map the four PBO resources.
    cudaGraphicsResource_t resources[] =
    {
        dispBuffer.pbo_resource
    };
    const size_t resourcesLen = sizeof resources / sizeof cudaGraphicsResource_t;

    cudaGraphicsMapResources(resourcesLen, resources);
    
    // Retrieve the four mapped buffers.
    size_t num_bytes;
    void *d_dst[resourcesLen];
    for (size_t i = 0; i < resourcesLen; i++)
        cudaGraphicsResourceGetMappedPointer(&d_dst[i], &num_bytes, resources[i]);

    cuMemcpy((CUdeviceptr)d_dst[0], cudaBuffer, dispBuffer.size);
    
    // Unmap the PBO resources.
    cudaGraphicsUnmapResources(resourcesLen, resources);

    glColor3f(1, 1, 1);
    glPushMatrix();
    
        // Bind the main buffer.
        glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, dispBuffer.pbo);
    
        //// Adjust the display.
    
        // Pan first, then scale.
        const GLfloat xTransMag = abs(transitionToValue(displayXOffset, GLfloat(xOffset), 0.1f));
        const GLfloat yTransMag = abs(transitionToValue(displayYOffset, GLfloat(yOffset), 0.1f));
        if (0.2f > xTransMag && 0.2f > yTransMag)
            transitionToValue(displayScaling, zoomScalar, 0.1f);
    
        // Set the scaling.
        glScalef(displayScaling, displayScaling, 1.f);
    
        // Set the offset.
        const float xOffsetScalar = -2.f * displayXOffset / float(dispWidth);
        const float yOffsetScalar = -2.f * displayYOffset / float(dispHeight);
        glTranslatef(xOffsetScalar, yOffsetScalar, 0.f);
    
        //// Display the main buffer.
    
        // load texture from PBO
        glBindTexture(GL_TEXTURE_2D, dispBuffer.tex);
        if (1 == dispBuffer.bytedepth)
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, dispBuffer.width, dispBuffer.height, GL_LUMINANCE, GL_UNSIGNED_BYTE, (void*)0);
        else if (2 == dispBuffer.bytedepth)
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, dispBuffer.width, dispBuffer.height, GL_LUMINANCE, GL_UNSIGNED_SHORT, (void*)0);
        else if (3 == dispBuffer.bytedepth)
            glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, dispBuffer.width, dispBuffer.height, GL_RGB, GL_UNSIGNED_BYTE, (void*)0);
    
        // Determine the "real" display coordinates.
        const GLfloat bufWidth = GLfloat(dispBuffer.width);
        const GLfloat bufHeight = GLfloat(dispBuffer.height);
    
        const GLfloat xScalar = bufWidth / dispWidth;
        const GLfloat yScalar = bufHeight / dispHeight;
    
        const GLfloat upper_left[]  = { -1.f * xScalar, 1.f * yScalar };
        const GLfloat lower_left[]  = { -1.f * xScalar, -1.f * yScalar };
        const GLfloat upper_right[] = { 1.f * xScalar,  1.f * yScalar };
        const GLfloat lower_right[] = { 1.f * xScalar,  -1.f * yScalar };
    
        // Setup the texture quadrant.
        glBegin(GL_QUADS);
        glTexCoord2f(0.0f, 0.0f);   glVertex2fv(upper_left);
        glTexCoord2f(0.0f, 1.0f);   glVertex2fv(lower_left);
        glTexCoord2f(1.0f, 1.0f);   glVertex2fv(lower_right);
        glTexCoord2f(1.0f, 0.0f);   glVertex2fv(upper_right);
        glEnd();
    
    glPopMatrix();
    
    //// Display the overlay text.
    const auto overlayTextCommitted = [](void)
    {
        const std::lock_guard<std::mutex> l (externallyAccessibleVariableMutex);
        return overlayText;
    }();

    if (overlayTextCommitted.size())
    {
        renderText(overlayTextCommitted, 2, 24, int(dispHeight) - 24, 0, 0, 0);
        renderText(overlayTextCommitted, 2, 25, int(dispHeight) - 25, 1, 1, 1);
    }

    //// Update the display surface.
    
    // Count a displayed buffer.
    {
        const std::lock_guard<std::mutex> l (externallyAccessibleVariableMutex);
        bufferDisplayRate.count();
    }

    // Swap the front and back buffers.
    glutSwapBuffers();
}

// Keyboard handlers.
void gfsdSpecialFunc (int k, int, int)
{
    switch (k)
    {
    case GLUT_KEY_F11:
        glutFullScreenToggle();
        break;

    case GLUT_KEY_UP:
        yOffset += 10;
        break;
    case GLUT_KEY_DOWN:
        yOffset -= 10;
        break;
    case GLUT_KEY_LEFT:
        xOffset -= 10;
        break;
    case GLUT_KEY_RIGHT:
        xOffset += 10;
        break;
    }
}
void gfsdKeyboardFunc (unsigned char k, int, int)
{
    if (keyCallback && keyCallback(k))
        return;

    switch (k)
    {
    case 'Q': case 'q':
    case VK_ESCAPE:
        glutLeaveMainLoop();
        break;

    case '-':
        zoomScalar = std::max( pow(1.1f, -50.f), zoomScalar / 1.1f);
        break;
    case '+':
        zoomScalar = std::min( pow(1.1f, 50.f), zoomScalar * 1.1f);
        break;

    case '1':
        // Zoom to 100%, and nullify the offsets.
        xOffset = 0;
        yOffset = 0;
        zoomScalar = 1.f;
        break;

    case 'F': case 'f':
        // Zoom to fit the display area, and nullify the offsets.
        xOffset = 0;
        yOffset = 0;

        // Select the most appropriate scaling.
        zoomScalar = std::min(dispWidth / GLfloat(dispBuffer.width), dispHeight / GLfloat(dispBuffer.height));
        zoomScalar = std::min( pow(1.1f, 50.f), std::max( pow(1.1f, -50.f), zoomScalar ) );
        break;
    }
}

// Mouse wheel handler.
void gfsdMouseWheelFunc (int wheel, int direction, int x, int y)
{
    zoomScalar = std::max(std::min( zoomScalar * pow(1.1f, (float)direction), pow(1.1f, 50.f) ), pow(1.1f, -50.f));
}

// Mouse button handler.
bool panDrag = false;
int xOffsetInit = 0, yOffsetInit = 0;
int xPosInit = 0, yPosInit = 0;
void gfsdMouseFunc (int button, int state, int x, int y)
{
    if (GLUT_DOWN == state)
    {
        // Enable the appropriate drag operation.
        panDrag = (GLUT_LEFT_BUTTON == button);

        xOffsetInit = xOffset;
        yOffsetInit = yOffset;
        xPosInit = x;
        yPosInit = y;

        smoothTransitionsEnabled = false;
    }
    else
        smoothTransitionsEnabled = true;
}

// Mouse drag handler.
void gfsdMotionFunc (int x, int y)
{
    const float xDelta = float(xPosInit - x);
    const float yDelta = float(y - yPosInit);
    if (abs(xDelta) < 1.f || abs(yDelta) < 1.f)
        return;
    
    if (panDrag)
    {
        xOffset = (xOffsetInit + int(xDelta / zoomScalar));
        yOffset = (yOffsetInit + int(yDelta / zoomScalar));
    }
}

// GLUT configurators.
void initOpenGLBuffers (unsigned int width, unsigned int height, unsigned int bytedepth)
{
    // delete old buffers
    dispBuffer.clean();

    // check for minimized window
    if ((width == 0) && (height == 0))
        return;

    // Allocate the buffers.
    dispBuffer.make(width, height, bytedepth);
}

#ifdef _WIN32
// This is specifically to enable the application to enable/disable vsync
typedef BOOL (WINAPI *PFNWGLSWAPINTERVALFARPROC)(int);

void setVSync (int interval)
{
    if (WGL_EXT_swap_control)
    {
        wglSwapIntervalEXT = (PFNWGLSWAPINTERVALFARPROC)wglGetProcAddress("wglSwapIntervalEXT");
        wglSwapIntervalEXT(interval);
    }
}
#endif

// Enter the main-loop of the BFDVPsimpleGLUT engine. This will return when the user closes the display
// window. Only on thread should call this method at any given time.
void enterBFDVPsimpleGLUT (int argc, char *argv[], const unsigned int width, const unsigned int height, const unsigned int bitdepth, const bool fullScreen)
{
    // Setup the OpenGL window.
    glutInit(&argc, argv);
    
    glutInitDisplayMode(GLUT_LUMINANCE | GLUT_DOUBLE | GLUT_MULTISAMPLE);
    glutInitWindowSize(width, height);
    glutInitWindowPosition(0, 0);
    glutCreateWindow(argv[0]);
    
    // Update the display at about 60Hz.
    glutTimerFunc(1000/60, gfsdTimerEventFunc, 0);
    glutDisplayFunc(gfsdDisplayFunc);
    glutReshapeFunc(gfsdReshapeFunc);

    glutSpecialFunc(gfsdSpecialFunc);
    glutKeyboardFunc(gfsdKeyboardFunc);
    glutMouseWheelFunc(gfsdMouseWheelFunc);
    glutMouseFunc(gfsdMouseFunc);
    glutMotionFunc(gfsdMotionFunc);

    glewInit();

    // Initialize our buffers.
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
    
    const unsigned int imagePixBytes = (bitdepth + 7) / 8;
    initOpenGLBuffers(width, height, imagePixBytes);

#ifdef _WIN32
    setVSync(0);
#endif
    
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

    if (fullScreen)
        glutFullScreen();

    // Reset the rate counter.
    {
        const std::lock_guard<std::mutex> externallyAccessibleVariableLock (externallyAccessibleVariableMutex);
        bufferDisplayRate.reset();
    }

    // The GL main loop.
    glutMainLoop();
}

// Exit the GLUT main-loop on demand, closing any windows.
void exitBFDVPsimpleGLUT (void)
{
    glutLeaveMainLoop();
}

// Setup a callback function to be called whenever an ASCII keyboard key is pressed. If this function
// returns true, the display will ignore the keypress internally. If false, the keypress may invoke
// some standard internal behavior.
void setBFDVPsimpleGLUTkeyCallback (BFDVPsimpleGLUTkeyCallback callback)
{
    keyCallback = callback;
}

// Pass a buffer to BFDVPsimpleGLUT for display (we target 60Hz display; some buffers may be dropped).
void setMainBFDVPsimpleGLUT (const CUdeviceptr displayBuf)
{
    cudaBufferAtm.store(displayBuf);
}

// Set the overlay text.
void setBFDVPsimpleGLUToverlayText (const char *const text)
{
    const std::lock_guard<std::mutex> externallyAccessibleVariableLock (externallyAccessibleVariableMutex);
    overlayText = text;
}

// Return the estimated rate in Hz at which frames are being displayed.
double getBFDVPsimpleGLUTdisplayRateAverage (void)
{
    const std::lock_guard<std::mutex> externallyAccessibleVariableLock (externallyAccessibleVariableMutex);
    return bufferDisplayRate.average_rate();
}
