/* FILE:        BayerDVP.cpp
 * DATE:        10/9/2014
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2014, BitFlow, Inc.
 * DESCRIPTION: A simple demonstration of the BFDVP API, integrating with OpenCV processing,
 *              to perform a live Bayer demosaic.
 */

#define NOMINMAX

// Local helper includes.
#include "BFDVPsimpleGLUT.h"
#include "SimpleTimers.hpp"

// CUDA processing methods.
#include <opencv2/gpu/gpu.hpp>
#include "QuadrantChannelGain.h"

// Standard includes.
#include <iostream>
#include <sstream>

#include <atomic>
#include <thread>
#include <mutex>

#include <chrono>

#include <vector>
#include <string>

// BitFlow includes.
#include <BFApi.h>
#include <BFErApi.h>
#include <BiApi.h>

// BFDVP includes.
#include <BFDVPApi.h>

// CUDA includes.
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <cuda_profiler_api.h>

// CUDA profiler definitions.
#define CUDA_PROFILER_ENABLED 1

#if CUDA_PROFILER_ENABLED
#   define CUDA_PROFILER_START() (cudaProfilerStart())
#   define CUDA_PROFILER_STOP()  (cudaProfilerStop())
#else
#   define CUDA_PROFILER_START() (CUDA_SUCCESS)
#   define CUDA_PROFILER_STOP()  (CUDA_SUCCESS)
#endif

// The number of system-device buffer pairs to allocate for acquisition and
// processing. There must be at least two acquisition buffer pairs, plus an
// extra processing buffer pair, for a minimum valid count of 3 buffers.
BFU32       vfgNumber               = ~BFU32(0);
BFU32       bufferPairCount         = 50;
BFU32       displayBufferCount      = bufferPairCount;

BFU32       imageWidth              = 0;
BFU32       imageHeight             = 0;
BFU32       imageBitDepth           = 0;
BFU32       imageByteSize           = 0;

bool        useDvpDma               = true;

bool        startInFullScreen       = false;
bool        showOverlay             = true;
bool        modeAutoCycleEnabled    = false;

static const BFU32 MS_MIN = 1;
static const BFU32 MS_MAX = 60 * 1000 * 5;

    // Cycle to the next mode after this many milliseconds.
BFU32       cyclePeriodMs           = 5000;
    // Control the mode speed, if applicable for the mode.
BFU32       replayPeriodMs          = 30;

enum BayerOrder
{
    BayerGRBG,
    BayerGBRG,
    BayerRGGB,
    BayerBGGR
};

BayerOrder                      bayerOrder = BayerGRBG;
bool                            highQualityDemosaic = true;

const static double             ChannelGainIncrement = 0.002;
const static double             ChannelGainEpsilon = ChannelGainIncrement / 10.;
double                          rGain = 0.5;
double                          gGain = 0.5;
double                          bGain = 0.5;

RateAverager<>                  processingRateAverager;
std::mutex                      processingRateAveragerMutex;

std::atomic<unsigned int>       activeMode (0);
std::atomic<char*>              modeNameStr (nullptr);

// Do whatever processing we want to for the acquired buffer.
std::atomic<bool>               exitThreadsNow  (false);
std::atomic<PBFDVPSysDevPair>   processBufferIn (nullptr);

// Method to perform buffer processing.
typedef std::vector<cv::gpu::GpuMat>::iterator DispImgIterator;

void live_process (DispImgIterator &dispImgIter, DispImgIterator const& dispImgBegin, DispImgIterator const& dispImgEnd)
{
    modeNameStr.store("Live");

    BFRC rc;

    // Retrieve the last acquired frame. For the purposes of this example, don't worry
    // whether or not we have previously processed this image.
    PBFDVPSysDevPair bufPairIn = processBufferIn.load();
    if (!bufPairIn)
    {
        std::this_thread::sleep_for( std::chrono::milliseconds(1) );
        return;
    }

    // DMA the buffer to the device (necessary, to do any CUDA processing).
    if (useDvpDma)
    {
        // DMA the buffer to the device (necessary, to do any CUDA processing).
        rc = BFDVPSys2Dev(bufPairIn, nullptr);
        if (BF_OK != rc)
        {
            std::cerr << "Sys2Dev transfer failed.\n";
            return;
        }
    }
    else
    {
        // Attempt to copy the system buffer to the device input buffer.
        if (CUDA_SUCCESS != cuMemcpyHtoD(bufPairIn->devBuffer, bufPairIn->sysBuffer, imageByteSize))
        {
            std::cerr << "cuMemcpyHtoD transfer failed.\n";
            return;
        }
    }
    
    // Select the display buffer.
    cv::gpu::GpuMat &outputImage = *dispImgIter;
    if (++dispImgIter == dispImgEnd)
        dispImgIter = dispImgBegin;

    // Setup the Bayer color order.
    int cvBayerOrder;
    QuadrantGains8 channelGains;
    switch (bayerOrder)
    {
    case BayerGRBG:
        cvBayerOrder = highQualityDemosaic ? cv::gpu::COLOR_BayerGR2RGB_MHT : cv::COLOR_BayerGR2RGB;
        fillQuadGains8(&channelGains, gGain, rGain, bGain, gGain);
        break;
    case BayerGBRG:
        cvBayerOrder = highQualityDemosaic ? cv::gpu::COLOR_BayerGB2RGB_MHT : cv::COLOR_BayerGB2RGB;
        fillQuadGains8(&channelGains, gGain, bGain, rGain, gGain);
        break;
    case BayerRGGB:
        cvBayerOrder = highQualityDemosaic ? cv::gpu::COLOR_BayerRG2RGB_MHT : cv::COLOR_BayerRG2RGB;
        fillQuadGains8(&channelGains, rGain, gGain, gGain, bGain);
        break;
    case BayerBGGR:
        cvBayerOrder = highQualityDemosaic ? cv::gpu::COLOR_BayerBG2RGB_MHT : cv::COLOR_BayerBG2RGB;
        fillQuadGains8(&channelGains, bGain, gGain, gGain, rGain);
        break;
    }

    // Apply the channel gains.
    if ( abs(rGain - 0.5) > ChannelGainEpsilon || abs(bGain - 0.5) > ChannelGainEpsilon || abs(gGain - 0.5) > ChannelGainEpsilon )
        quadrantChannelGain8((std::uint8_t*)bufPairIn->devBuffer, (std::uint8_t*)bufPairIn->devBuffer, imageWidth, imageHeight, channelGains);

    // Wrap the input image for OpenCV.
    cv::gpu::GpuMat inputImage (imageWidth, imageHeight, CV_8UC1, (void*)bufPairIn->devBuffer);
    
    // Demosaic the image.
    cv::gpu::demosaicing(inputImage, outputImage, cvBayerOrder, 3);
    
    // Display the image.
    setMainBFDVPsimpleGLUT( (CUdeviceptr)outputImage.ptr() );

    // Iterate the processing rate averager.
    {
        const std::lock_guard<std::mutex> processingRateAveragerLock (processingRateAveragerMutex);
        processingRateAverager.count();
    }
}

// Slow-mo playback.
void slowmo_playback (DispImgIterator &dispImgIter, DispImgIterator const& dispImgBegin, DispImgIterator const& dispImgEnd)
{
    modeNameStr.store("Replay");

    // Iterate the processing rate averager, but only if enough time has passed.
    bool drawNext = false;
    {
        const std::lock_guard<std::mutex> processingRateAveragerLock (processingRateAveragerMutex);
        if (drawNext = (processingRateAverager.active_period() >= replayPeriodMs))
            processingRateAverager.count();
    }

    // If sufficient time has passed, draw the next image.
    if (drawNext)
    {
        // Select the display buffer.
        cv::gpu::GpuMat &outputImage = *dispImgIter;
        if (++dispImgIter == dispImgEnd)
            dispImgIter = dispImgBegin;

        // Display the image.
        setMainBFDVPsimpleGLUT( (CUdeviceptr)outputImage.ptr() );
    }
    else
        std::this_thread::yield();
}

// Thread function for buffer processing.
void buffer_process_thread (void)
{
    // Mode variables.
    auto cycleStartTime = std::chrono::high_resolution_clock::now();
    auto modeSpeedTime = cycleStartTime;

    // Setup the display buffers.
    std::vector<cv::gpu::GpuMat> cvGpuDispImages (displayBufferCount);
    for (auto &gpuImg : cvGpuDispImages)
        gpuImg = cv::gpu::GpuMat(imageWidth, imageHeight, CV_8UC3);
    const auto dispImgBegin = cvGpuDispImages.begin(), dispImgEnd = cvGpuDispImages.end();
    auto dispImgIter = dispImgBegin;

    while (!exitThreadsNow.load())
    {
        // Select the mode.
        if (modeAutoCycleEnabled)
        {
            const auto timeNow = std::chrono::high_resolution_clock::now();
            if (timeNow - cycleStartTime >= std::chrono::milliseconds(cyclePeriodMs))
            {
                activeMode++;
                cycleStartTime = timeNow;
                
                const std::lock_guard<std::mutex> processingRateAveragerLock (processingRateAveragerMutex);
                processingRateAverager.reset();
            }
        }

        // Handle processing for the active mode.
        switch (activeMode)
        {
        default:
            activeMode = 0;
        case 0:
            live_process(dispImgIter, dispImgBegin, dispImgEnd);
            break;
        case 1:
            slowmo_playback(dispImgIter, dispImgBegin, dispImgEnd);
            break;
        }
    }
}

// The acquisition thread loop.
void enterAcquisitionMainLoop ( const Bd hBoard
                              , const PBIBA pBufArray
                              , PBFDVPSysDevPair sysDevPairs )
{
    BFRC rc;

    // Loop acquisition until the user presses a key, or the acquisition aborts.
    BFU32 timeoutCnt = 0;
    BiCirHandle cirHandle;

    // Setup counters.
    {
        const std::lock_guard<std::mutex> processingRateAveragerLock (processingRateAveragerMutex);
        processingRateAverager.reset(2 * bufferPairCount);
    }
    RateAverager<> acquisitionRateAverager (2 * bufferPairCount);
    PassiveTimer<> dispUpdateTimer (1000 / 15);

    // Loop indefinitely, processing every frame as it becomes available.
    while (!exitThreadsNow.load())
    {
        // Print a status message at about 60Hz (or less).
        if (dispUpdateTimer.timeout())
        {
            size_t procdFrmCnt;
            double procdRteAvg;
            {
                const std::lock_guard<std::mutex> processingRateAveragerLock (processingRateAveragerMutex);
                procdFrmCnt = processingRateAverager.total_periods();
                procdRteAvg = processingRateAverager.average_rate();
            }
            
            const size_t acqdFrmCnt = acquisitionRateAverager.total_periods();
            const double acqdRteAvg = acquisitionRateAverager.average_rate();

            std::cout << acqdFrmCnt << " acquired, " << procdFrmCnt << " processed, " << timeoutCnt << " timeouts\r";

            if (showOverlay)
            {
                const char *const modeName = modeNameStr.load();
                char overlayText [512];
                sprintf(overlayText, "Proc. Mode: %s %s\nProc. Rate: %6.2f FPS\nAcqu. Rate: %6.2f FPS\nDisp. Rate: %6.2f FPS"
                    , modeName ? modeName : "Unknown :-("
                    , modeAutoCycleEnabled ? "[auto]" : "[manual]"
                    , procdRteAvg
                    , acqdRteAvg
                    , getBFDVPsimpleGLUTdisplayRateAverage() );

                setBFDVPsimpleGLUToverlayText(overlayText);
            }
            else
                setBFDVPsimpleGLUToverlayText("");
        }

        // Wait for a frame.
        rc = BiCirWaitDoneFrame(hBoard, pBufArray, 100, &cirHandle);
        switch (rc)
        {
        case BF_OK:
            processBufferIn.store( &sysDevPairs[cirHandle.BufferNumber] );

            // Make the buffer available to the acquisition engine again.
            acquisitionRateAverager.count();
            BiCirStatusSet(hBoard, pBufArray, cirHandle, BIAVAILABLE);
            continue;

        case BI_ERROR_CIR_WAIT_TIMEOUT:
            timeoutCnt++;
            continue;
        }

        // At any other error, quit acquisition. Abort indicates a successful exit.
        if (BI_CIR_ABORTED != rc)
            BiErrorShow(hBoard, rc);
        break;
    }
}

// External key-press handler for the BFDVPsimpleGLUT display.
bool keyPressHandler (const char key)
{
    switch (key)
    {
    case 'l': case 'L':
        highQualityDemosaic = false;
        break;
    case 'h': case 'H':
        highQualityDemosaic = true;
        break;

    case 'o': case 'O':
        showOverlay = !showOverlay;
        std::cout << "\n" << (showOverlay ? "Enabled" : "Disabled") << " the text overlay.\n";
        return true;

    case 'd': case 'D':
        useDvpDma = !useDvpDma;
        std::cout << "\n" << (useDvpDma ? "Enabled" : "Disabled") << " DVP DMA.\n";
        return true;

    case '!': // Shift+1
        bayerOrder = BayerGRBG;
        break;
    case '@': // Shift+2
        bayerOrder = BayerGBRG;
        break;
    case '#': // Shift+3
        bayerOrder = BayerRGGB;
        break;
    case '$': // Shift+4
        bayerOrder = BayerBGGR;
        break;

    case 'r':
        rGain -= ChannelGainIncrement;
        break;
    case 'R':
        rGain += ChannelGainIncrement;
        break;
    case 'g':
        gGain -= ChannelGainIncrement;
        break;
    case 'G':
        gGain += ChannelGainIncrement;
        break;
    case 'b':
        bGain -= ChannelGainIncrement;
        break;
    case 'B':
        bGain += ChannelGainIncrement;
        break;

    case '\t':
        modeAutoCycleEnabled = false;
        activeMode++;
        break;
    case 'c':
        cyclePeriodMs = std::max<BFU32>(MS_MIN, cyclePeriodMs - 1);
        break;
    case 'C':
        cyclePeriodMs = std::min<BFU32>(MS_MAX, cyclePeriodMs + 1);
        break;
    case 's':
        replayPeriodMs = std::min<BFU32>(MS_MAX, replayPeriodMs + 1);
        break;
    case 'S':
        replayPeriodMs = std::max<BFU32>(MS_MIN, replayPeriodMs - 1);
        break;
    case 'a': case 'A':
        modeAutoCycleEnabled = !modeAutoCycleEnabled;
        break;

    default:
        return false;
    }

    // Validate the gains.
    rGain = clampLinGain(rGain);
    gGain = clampLinGain(gGain);
    bGain = clampLinGain(bGain);

    // Display the current status.
    std::cout << "\n" << (highQualityDemosaic ? "Malvar-He-Cutler" : "Bilinear") << " demosaic ";
    switch (bayerOrder)
    {
    case BayerGRBG:
        std::cout << "GRBG";
        break;
    case BayerGBRG:
        std::cout << "GBRG";
        break;
    case BayerRGGB:
        std::cout << "RGGB";
        break;
    case BayerBGGR:
        std::cout << "BGGR";
        break;
    }
    std::cout << " with RGB Gains = { " << rGain << ", " << gGain << ", " << bGain << " }\n";

    std::cout << "Mode ";
    if (modeAutoCycleEnabled)
        std::cout << "auto-cycling with " << cyclePeriodMs << "ms period and ";
    else
        std::cout << "index " << activeMode << ", ";
    std::cout << (1000. / replayPeriodMs) << " FPS replay speed.\n";

    return true;
}

typedef std::vector<std::string> ArgsVector;
bool processArgs (ArgsVector const& args)
{
    unsigned int uiTmp;
    std::string badArgument;

    for (auto &arg : args)
    {
        if (1 == sscanf(arg.c_str(), "--board=%u", &vfgNumber))
            std::cout << "Selected BitFlow board index " << vfgNumber << ".\n";
        else if (1 == sscanf(arg.c_str(), "--acqBufs=%u", &bufferPairCount))
        {
            std::cout << "Selected " << bufferPairCount << " acquisition buffers.\n";
            if (2 > bufferPairCount)
            {
                std::cerr << "ERROR: At least two buffers are required for acquisition!\n";
                badArgument = arg;
                break;
            }
        }
        else if (1 == sscanf(arg.c_str(), "--dispBufs=%u", &displayBufferCount))
        {
            std::cout << "Selected " << displayBufferCount << " display buffers.\n";
            if (1 > displayBufferCount)
            {
                std::cerr << "Error: At least one buffer is required for display!\n";
                badArgument = arg;
                break;
            }
        }
        else if (arg == "--GRBG")
        {
            std::cout << "Selected the GRBG Bayer pattern.\n";
            bayerOrder = BayerGRBG;
        }
        else if (arg == "--GBRG")
        {
            std::cout << "Selected the GBRG Bayer pattern.\n";
            bayerOrder = BayerGBRG;
        }
        else if (arg == "--RGGB")
        {
            std::cout << "Selected the RGGB Bayer pattern.\n";
            bayerOrder = BayerRGGB;
        }
        else if (arg == "--BGGR")
        {
            std::cout << "Selected the BGGR Bayer pattern.\n";
            bayerOrder = BayerBGGR;
        }
        else if (arg == "--bilinear")
        {
            std::cout << "Selected bilinear demosaic.\n";
            highQualityDemosaic = false;
        }
        else if (arg == "--malvar")
        {
            std::cout << "Selected Malvar-He-Cutler demosaic.\n";
            highQualityDemosaic = true;
        }
        else if (3 == sscanf(arg.c_str(), "--gain=%lf:%lf:%lf", &rGain, &gGain, &bGain))
        {
            std::cout << "Selected RGB gain = { " << rGain << ", " << gGain << ", " << bGain << " }\n";

            if (abs(clampLinGain(rGain) - rGain) > ChannelGainEpsilon)
            {
                std::cerr << "ERROR: the specified red channel gain is not in the range [0.0, 1.0].\n";
                badArgument = arg;
                break;
            }

            if (abs(clampLinGain(gGain) - gGain) > ChannelGainEpsilon)
            {
                std::cerr << "ERROR: the specified green channel gain is not in the range [0.0, 1.0].\n";
                badArgument = arg;
                break;
            }

            if (abs(clampLinGain(bGain) - bGain) > ChannelGainEpsilon)
            {
                std::cerr << "ERROR: the specified blue channel gain is not in the range [0.0, 1.0].\n";
                badArgument = arg;
                break;
            }
        }
        else if (1 == sscanf(arg.c_str(), "--cyclePrd=%u", &cyclePeriodMs))
        {
            std::cout << "Selected mode auto-cycle period = " << cyclePeriodMs << "ms.\n";
            if (cyclePeriodMs > MS_MAX || cyclePeriodMs < MS_MIN)
            {
                std::cerr << "ERROR: the specified mode auto-cycle millisecond period is not in the range [10ms, 5min].\n";
                badArgument = arg;
                break;
            }
        }
        else if (1 == sscanf(arg.c_str(), "--replayPrd=%u", &replayPeriodMs))
        {
            std::cout << "Selected replay period = " << replayPeriodMs << "ms (" << (1000. / replayPeriodMs) << " FPS).\n";
            if (replayPeriodMs > MS_MAX || replayPeriodMs < MS_MIN)
            {
                std::cerr << "ERROR: the specified replay millisecond period is not in the range [10ms, 5min].\n";
                badArgument = arg;
                break;
            }
        }
        else if (1 == sscanf(arg.c_str(), "--mode=%u", &uiTmp))
        {
            std::cout << "Selected the " << uiTmp << " mode index.\n";
            activeMode = uiTmp;
        }
        else if (arg == "--autoCycle")
        {
            std::cout << "The mode will auto-cycle.\n";
            modeAutoCycleEnabled = true;
        }
        else if (arg == "--fullscreen")
        {
            std::cout << "The application will start in full-screen mode.\n";
            startInFullScreen = true;
        }
        else if (arg == "--noOverlay")
        {
            std::cout << "Disabled the overlay.\n";
            showOverlay = false;
        }
        else if (arg == "--noDvp")
        {
            std::cout << "Disabled DVP DMA.\n";
            useDvpDma = false;
        }
        else
        {
            std::cerr << "Unknown argument!\n";
            badArgument = arg;
            break;
        }
    }

    if (badArgument.size())
    {
        if (badArgument != "?" && badArgument != "/?" && badArgument != "h" && badArgument != "-h" && badArgument != "--help" && badArgument != "help")
            std::cerr << "ERROR: Found a bad input argument: " << badArgument << "\n";

        /* 80 cols:   cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc */
        std::cerr << "Known arguments are: \n"
                     "  --board=N     : Set the BitFlow board to be opened (default asks).\n"
                     "  --acqBufs=N   : The Number of acquisition buffers (def 50, min 2).\n"
                     "  --dispBufs=N  : The Number of display buffers (def 50, min 1).\n"
                     "  --GRBG        : Select the specified Bayer pattern.\n"
                     "  --GBRG        : Select the specified Bayer pattern.\n"
                     "  --RGGB        : Select the specified Bayer pattern.\n"
                     "  --BGGR        : Select the specified Bayer pattern.\n"
                     "  --bilinear    : Select bilinear (low-quality) Bayer demosaic.\n"
                     "  --malvar      : Select Malvar-He-Cutler (high-quality) Bayer demosaic.\n"
                     "  --gain=R:G:B  : The Red, Green and Blue channel gains, from [0.0, 1.0].\n"
                     "  --cyclePrd=P  : Mode auto-cycle period in ms (def 5000, rng [" << MS_MIN << ", " << MS_MAX << "]).\n"
                     "  --replayPrd=P : Replay frame-period in ms (def 30, rng [" << MS_MIN << ", " << MS_MAX << "]).\n"
                     "  --mode=N      : Set the startup mode to index N (def 0).\n"
                     "  --autoCycle   : Enable mode auto-cycling.\n"
                     "  --noDvp       : Disable DVP DMA at startup.\n"
                     "  --fullscreen  : Set the application window to fullscreen mode.\n"
                     "  --noOverlay   : Disable the text overlay.\n";
        return false;
    }

    return true;
}

// The main method.
int main (const int argc, char *argv[])
{
    BFRC rc;

    // Process any arguments.
    const ArgsVector args (argv + 1, argv + argc);
    if (!processArgs(args))
        return EXIT_FAILURE;
    
    // Open the board.
    Bd hBoard;
    if ( !(~vfgNumber) )
    {
        // Query the user for a board to open.
        BFU32 BoardType, BoardNum;
        BFU32 Init, SerNum;
        if ( IDCANCEL == DoBrdOpenDialog(BOD_HIDEJUSTOPEN, FF_BITFLOW_MODERN, &BoardType, &BoardNum, &Init, &SerNum) )
            return EXIT_SUCCESS;
    
        rc = BiBrdOpen(BoardType, BoardNum, &hBoard);
        if (BF_OK != rc)
        {
            std::cerr << "Failed to open BitFlow board " << BoardNum << ".\n";
            return EXIT_FAILURE;
        }
    }
    else
    {
        rc = BiBrdOpen(BiTypeAny, vfgNumber, &hBoard);
        if (BF_OK != rc)
        {
            std::cerr << "Failed to open BitFlow board " << vfgNumber << ".\n";
            return EXIT_FAILURE;
        }
    }

    // Retrieve the requisite board parameters.
    BiBrdInquire(hBoard, BiCamInqXSize, &imageWidth);
    BiBrdInquire(hBoard, BiCamInqYSize0, &imageHeight);
    BiBrdInquire(hBoard, CiCamInqPixBitDepth, &imageBitDepth);

    // Verify that we support the input bitdepth.
    switch (imageBitDepth)
    {
    case 8:
        break;

    default:
        // Unwind
        BiBrdClose(hBoard);

        std::cerr << "Bit-depth " << imageBitDepth << " unsupported by this demo.\n";
        return EXIT_FAILURE;
    }
    
    const BFU32 imageByteDepth = (imageBitDepth + 7) / 8;
    imageByteSize = imageWidth * imageHeight * imageByteDepth;
    
    // At application exit, call cudaDeviceReset.
    atexit([]() { cudaDeviceReset(); });
    
    // Setup the CUDA context.
    if (CUDA_SUCCESS != CUDA_PROFILER_START())
    {
        BiBrdClose(hBoard);

        std::cerr << "Unable to start the CUDA profiler.\n";
        return EXIT_FAILURE;
    }

    if (CUDA_SUCCESS != cuInit(0))
    {
        // Unwind
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Unable to initialize the CUDA library.\n";
        return EXIT_FAILURE;
    }

    CUcontext cudaCtx;
    if (CUDA_SUCCESS != cuCtxCreate(&cudaCtx, 0, 0))
    {
        // Unwind
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to create the CUDA context object.";
        return EXIT_FAILURE;
    }

    if (CUDA_SUCCESS != cuCtxSetCurrent(cudaCtx))
    {
        // Unwind
        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to set the current CUDA context.";
        return EXIT_FAILURE;
    }

    // Initialize the DVP context.
    rc = BFDVPInit(BFDVPInit_Default);
    if (BF_OK != rc)
    {
        // Unwind
        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to initialize the DVP context.\n";
        return EXIT_FAILURE;
    }
    
    // Allocate system and GPU device buffers using the BFDVPAllocAndBindPairs method.
    std::vector<BFDVPSysBuf>        systemBuffers   (bufferPairCount);
    std::vector<BFDVPSysDevPair>    sysDevPairs     (bufferPairCount);
    rc = BFDVPAllocAndBindPairs(systemBuffers.data(), sysDevPairs.data(), imageWidth, imageHeight, imageBitDepth, bufferPairCount, nullptr);
    if (BF_OK != rc)
    {
        // Unwind
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to allocate and bind the system and device buffer pairs.\n";
        return EXIT_FAILURE;
    }

    // Assign the allocated system buffers to the BitFlow acquisition engine. The final
    // buffer is reserved for processing, and therefore excluded from the acquisition engine.
    BIBA bufArray;
    rc = BiBufferAssign(hBoard, &bufArray, systemBuffers.data(), bufferPairCount - 1);
    if (BF_OK != rc)
    {
        // Unwind
        BFDVPBreakAndFreePairs(sysDevPairs.data(), bufferPairCount);
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to assign the system buffers to the BitFlow acquisition engine.\n";
        return EXIT_FAILURE;
    }

    // Configure a circular acquisition.
    rc = BiCircAqSetup(hBoard, &bufArray, CirErStop, BiAqEngJ);
    if (BF_OK != rc)
    {
        // Unwind
        BiBufferUnassign(hBoard, &bufArray);
        BFDVPBreakAndFreePairs(sysDevPairs.data(), bufferPairCount);
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to setup the frame grabber for a circular acquisition.\n";
        return EXIT_FAILURE;
    }
    
    /*   cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc */
    std::cout <<
        "Keyboard controls:\n"
        "  [Q|Esc]      : Exit the program.\n"
        "  [+/-]        : Zoom in/out.\n"
        "  [1]          : Zoom to 100%, and reset panning.\n"
        "  [F]          : Zoom to Fit the display window, and reset panning.\n"
        "  [r/R]        : Decrease/increase the Red channel gain.\n"
        "  [g/G]        : Decrease/increase the Green channel gain.\n"
        "  [b/B]        : Decrease/increase the Blue channel gain.\n"
        "  [L/H]        : Select Low/High quality Bayer (bilinear/Malvar-He-Cutler)\n"
        "  [!/@/#/$]    : Set the Bayer filter (GRBG/GBRG/RGGB/BGGR).\n"
        "  [Tab]        : Cycle to the next mode.\n"
        "  [c/C]        : Decrease/increase the mode auto-Cycling period.\n"
        "  [s/S]        : Decrease/increase the replay Speed.\n"
        "  [A]          : Toggle mode Auto-cycling.\n"
        "  [O]          : Toggle visibility of the text Overlay.\n"
        "  [F11]        : Toggle fullscreen display.\n"
        "Mouse controls:\n"
        "  Scroll-wheel : Zoom in or out.\n"
        "  Right-click  : Press, hold and drag, to pan the display.\n\n"
        "Closing the display window exits.\n\n";

    std::thread glutThread, acqThread;

    // Start acquisition.
    rc = BiCirControl(hBoard, &bufArray, BISTART, BiAsync);
    if (BF_OK == rc)
    {
        // Setup the display key-press handler.
        setBFDVPsimpleGLUTkeyCallback(&keyPressHandler);

        // Enter the BFDVPsimpleGLUT main loop.
        glutThread = std::thread( [&](void)
        {
            enterBFDVPsimpleGLUT(argc, argv, imageWidth, imageHeight, 3 * imageBitDepth, startInFullScreen);

            // Abort acquisition and processing.
            exitThreadsNow.store(true);
            BiCirControl(hBoard, &bufArray, BIABORT, BiAsync);
        } );

        // Enter the acquisition main loop.
        acqThread = std::thread( &enterAcquisitionMainLoop, hBoard, &bufArray, sysDevPairs.data() );

        // Enter the processing loop. CUDA operations need to take place in _this_ thread.
        buffer_process_thread();
    }

    // Cleanup everything.
    if (glutThread.joinable())
        glutThread.join();
    if (acqThread.joinable())
        acqThread.join();
    
    BiCircCleanUp(hBoard, &bufArray);
    BiBufferUnassign(hBoard, &bufArray);
    BFDVPBreakAndFreePairs(sysDevPairs.data(), bufferPairCount);
    BFDVPDeinit();

    cuCtxDestroy(cudaCtx);
    CUDA_PROFILER_STOP();
    BiBrdClose(hBoard);

    // All done!
    return EXIT_SUCCESS;
}
