#include "QuadrantChannelGain.h"
#include "CudaHelpers.h"

#include "device_launch_parameters.h"
#include "cuda_runtime_api.h"

inline __device__ void quadrantGain8 (const std::uint16_t in, std::uint8_t *const out, const std::uint16_t gNumerator)
{
    const std::uint16_t gDenominator = 256;
    const std::uint16_t gainApplied = in * gNumerator / gDenominator;
    *out = std::uint8_t(gainApplied >= gDenominator ? gDenominator - 1 : gainApplied);
}
inline __device__ void quadrantGainWord8 (const uchar4 in, uchar4 *const out, const PairGains8 pairGains)
{
    quadrantGain8(in.w, &out->w, pairGains.x[0]);
    quadrantGain8(in.x, &out->x, pairGains.x[1]);
    quadrantGain8(in.y, &out->y, pairGains.x[0]);
    quadrantGain8(in.z, &out->z, pairGains.x[1]);
}
inline __device__ void quadrantGainQWord8 (const uint4 in, uint4 *const out, const PairGains8 pairGains)
{
    quadrantGainWord8( *reinterpret_cast<const uchar4*>(&in.w), reinterpret_cast<uchar4*>(&out->w), pairGains );
    quadrantGainWord8( *reinterpret_cast<const uchar4*>(&in.x), reinterpret_cast<uchar4*>(&out->x), pairGains );
    quadrantGainWord8( *reinterpret_cast<const uchar4*>(&in.y), reinterpret_cast<uchar4*>(&out->y), pairGains );
    quadrantGainWord8( *reinterpret_cast<const uchar4*>(&in.z), reinterpret_cast<uchar4*>(&out->z), pairGains );
}

__global__ void doQuadrantChannelGain8 (const uint4 *const inBuf, uint4 *const outBuf, const std::uint32_t qwordWidth, const dim3 blockqwordsize, const std::uint32_t threadqwordsize, const QuadrantGains8 quadrantGains)
{
    std::uint32_t offset = blockIdx.y * blockqwordsize.y + blockIdx.x * blockqwordsize.x + threadIdx.x * threadqwordsize;

    const uint4 *in = inBuf + offset;
    uint4 *out = outBuf + offset;

    std::uint32_t qwordsRemaining = threadqwordsize;
    do
    {
        const std::uint32_t line = offset / qwordWidth;
        quadrantGainQWord8(*in, out, quadrantGains.y[1u & line]);
        ++in; ++out; ++offset;
    }
    while (--qwordsRemaining);
}

// Per-element image processing kernel, multiplying each input pixel by it's
// matching quadrant gain, then dividing by 255. The image width and height
// must be multiples of 16 bytes.
extern "C" void quadrantChannelGain8 ( const std::uint8_t *const devImgIn
                                     , std::uint8_t *const devImgOut
                                     , const std::uint32_t width
                                     , const std::uint32_t height
                                     , const QuadrantGains8 quadrantGains )
{
    uint32_t qwordsdown, threadcnt, threadsize;
    dim3 blockcnt, blocksize;

    const uint4 *inBuf  = reinterpret_cast<const uint4*>(devImgIn);
    uint4       *outBuf = reinterpret_cast<uint4*>(devImgOut);

    const std::uint32_t byteSize = width * height;
    const std::uint32_t qwordSize = byteSize / 4 / 4;
    const std::uint32_t qwordWidth = width / 4 / 4;
    
    // Process as many QWORDs simultaneously as supported by the device.
    std::uint32_t qwordsRemaining = qwordSize;
    bufferDiffLaunchVariablesQWord(qwordsRemaining, &qwordsdown, &blockcnt, &threadcnt, &blocksize, &threadsize);
    while (threadsize)
    {
        doQuadrantChannelGain8<<<blockcnt, threadcnt>>>( inBuf, outBuf, qwordWidth, blocksize, threadsize, quadrantGains );
        inBuf += qwordsdown;
        outBuf += qwordsdown;
        qwordsRemaining -= qwordsdown;
        
        // Process any additional QWORDs.
        bufferDiffLaunchVariablesQWord(qwordsRemaining, &qwordsdown, &blockcnt, &threadcnt, &blocksize, &threadsize);
    }
}
