/*
 * Copyright 1993-2013 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */

/*
    Recursive Gaussian filter
*/

#ifndef _RECURSIVEGAUSSIAN_KERNEL_CU_
#define _RECURSIVEGAUSSIAN_KERNEL_CU_

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "helper_cuda.h"
#include "helper_math.h"

#include "RecursiveGaussianBlur.h"

#define BLOCK_DIM           16
#define CLAMP_TO_EDGE       0

#define INT_ACCURACY_NUM    0xFFFFFF
#define INT_ACCURACY_DNM    0xFF
#define INT_ACCURACY        (INT_ACCURACY_NUM / INT_ACCURACY_DNM)
static_assert(INT_ACCURACY_NUM % INT_ACCURACY_DNM == 0, "The INT_ACCURACY NUM and DNM don't divide evenly.");

#define FLT2INT(FLT)        int(FLT * float(INT_ACCURACY))
#define CHR2INT(CHR)        (int(CHR) * INT_ACCURACY)
#define INT2CHR(INT)        (INT / INT_ACCURACY)

enum
{
    INT_SIGNBIT = sizeof(int) * 8 - 1
};

// Transpose kernel (see transpose SDK sample for details)
__global__ void d_transpose(uchar *odata, uchar *idata, int width, int height)
{
    __shared__ uchar block[BLOCK_DIM][BLOCK_DIM+1];

    // read the matrix tile into shared memory
    unsigned int xIndex = blockIdx.x * BLOCK_DIM + threadIdx.x;
    unsigned int yIndex = blockIdx.y * BLOCK_DIM + threadIdx.y;

    if ((xIndex < width) && (yIndex < height))
    {
        unsigned int index_in = yIndex * width + xIndex;
        block[threadIdx.y][threadIdx.x] = idata[index_in];
    }

    __syncthreads();

    // write the transposed matrix tile to global memory
    xIndex = blockIdx.y * BLOCK_DIM + threadIdx.x;
    yIndex = blockIdx.x * BLOCK_DIM + threadIdx.y;

    if ((xIndex < height) && (yIndex < width))
    {
        unsigned int index_out = yIndex * height + xIndex;
        odata[index_out] = block[threadIdx.x][threadIdx.y];
    }
}

// Clamp the value to the range [0, MAX].
template<unsigned int MAX>
__device__ int clampToMax (const int val)
{
    // To eliminate branching, use some bit-twidling.
    const unsigned int uval = reinterpret_cast<const unsigned int&>(val);
    const unsigned int uvalIsNgtv = uval >> INT_SIGNBIT;
    const unsigned int uvalNtNgtv = 1u - uvalIsNgtv;
    const unsigned int uvalInUppr = (uval - MAX - 1u) >> INT_SIGNBIT;
    const unsigned int uvalOvrUpr = 1u - uvalInUppr;
    const unsigned int uvalIsGrtr = uvalNtNgtv & uvalOvrUpr;
    const unsigned int uvalInRang = uvalNtNgtv & uvalInUppr;
    return uval * uvalInRang + MAX * uvalIsGrtr;
}

#define clampToUChar(VAL)   clampToMax<0xFF>(VAL)
#define clampToInt(VAL)     clampToMax<0xFF * INT_ACCURACY>(VAL)

/*
    recursive Gaussian filter

    parameters:
    id - pointer to input data (mono image packed into 8-bit integers)
    od - pointer to output data
    w  - image width
    h  - image height
    a0-a3, b1, b2, coefp, coefn - filter parameters
*/

__global__ void
d_recursiveGaussian_mono(uchar *id, uchar *od, int w, int h, int a0, int a1, int a2, int a3, int b1, int b2, int coefp, int coefn)
{
    unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;

    if (x >= w)
        return;

    id += x;    // advance pointers to correct column
    od += x;

    // forward pass
    int xp = 0;  // previous input
    int yp = 0;  // previous output
    int yb = 0;  // previous output by 2
#if CLAMP_TO_EDGE
    xp = *id;
    yb = INT2CHR(clampToInt(coefp*xp));
    yp = yb;
#endif
    
    for (int y = 0; y < h; y++)
    {
        int xc = *id;
        int yc = clampToInt(a0*xc + a1*xp - b1*yp - b2*yb);
        id += w;
        od += w;    // move to next row
        xp = xc;
        yb = yp;
        yp = *od = INT2CHR(yc);
    }

    // reset pointers to point to last element in column
    id -= w;
    od -= w;

    // reverse pass
    // ensures response is symmetrical
    int xn = 0;
    int xa = 0;
    int yn = 0;
    int ya = 0;
#if CLAMP_TO_EDGE
    xn = xa = *id;
    yn = INT2CHR(clampToInt(coefn*xn));
    ya = yn;
#endif
    
    for (int y = h-1; y >= 0; y--)
    {
        int xc = *id;
        int yc = clampToInt(a2*xn + a3*xa - b1*yn - b2*ya);
        xa = xn;
        xn = xc;
        ya = yn;
        yn = INT2CHR(yc);
        *od = clampToUChar(*od + INT2CHR(yc));
        id -= w;
        od -= w;  // move to previous row
    }
}

#endif // #ifndef _GAUSSIAN_KERNEL_H_
