#ifndef INCLUDED__SIMPLE__TIMERS__HPP
#define INCLUDED__SIMPLE__TIMERS__HPP

/* FILE:        SimpleTimers.hpp
 * DATE:        10/9/2014
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2014, BitFlow, Inc.
 * DESCRIPTION: A set of convenience classes related to timing.
 */

#include <vector>
#include <chrono>
#include <algorithm>

// A timer that only times-out when explicitly checked by the user.
template<typename ClockT = std::chrono::high_resolution_clock>
class PassiveTimer
{
public:
    typedef ClockT Clock;

private:
    // Illegal.
    PassiveTimer (void);
    PassiveTimer (PassiveTimer const&);
    PassiveTimer& operator= (PassiveTimer const&);

    // Members.
    const typename Clock::duration  m_period;
    typename Clock::time_point      m_lastTimeout;

public:
    PassiveTimer (typename Clock::duration const& d_period)
        : m_period      (d_period)
        , m_lastTimeout (Clock::now())
    { }
    PassiveTimer (unsigned int const& ms_period)
        : m_period      ( std::chrono::milliseconds(ms_period) )
        , m_lastTimeout (Clock::now())
    { }

    inline typename Clock::duration period (void) const
    {
        return m_period;
    }

    inline void reset (void)
    {
        m_lastTimeout = Clock::now();
    }

    inline bool peek (void) const
    {
        return Clock::now() - m_lastTimeout >= m_period;
    }

    inline bool timeout (void)
    {
        const auto now = Clock::now();
        if (now - m_lastTimeout < m_period)
            return false;
        
        m_lastTimeout = now;
        return true;
    }
};

// Estimate the average rate at which the count method is invoked.
template<typename ClockT = std::chrono::high_resolution_clock>
class RateAverager
{
public:
    typedef ClockT                      Clock;
    typedef typename Clock::duration    Duration;
    typedef typename Clock::time_point  TimePoint;

private:
    // Illegal.
    RateAverager (RateAverager const&);
    RateAverager& operator= (RateAverager const&);

    // Members.
    typedef std::vector<Duration>                   PeriodVector;
    typedef typename PeriodVector::iterator         PeriodIter;
    typedef typename PeriodVector::const_iterator   PeriodConstIter;

    PeriodVector    m_periods;
    PeriodConstIter m_firstPeriod;
    PeriodIter      m_nextPeriod;

    TimePoint       m_lastCountTime;
    size_t          m_totalPeriods;

public:
    RateAverager (const size_t prevoiusPeriods = 50)
        : m_periods         (prevoiusPeriods + 1)
        , m_lastCountTime   (Clock::now())
        , m_totalPeriods    (0)
    {
        m_firstPeriod = m_periods.begin();
        m_nextPeriod = m_periods.begin();
    }

    // Reset the period count, optionally setting a new period count.
    inline void reset (const size_t prevoiusPeriods = ~size_t(0))
    {
        if (~prevoiusPeriods)
            m_periods.resize(prevoiusPeriods + 1);

        m_lastCountTime = Clock::now();
        m_totalPeriods = 0;

        m_firstPeriod = m_periods.begin();
        m_nextPeriod = m_periods.begin();
    }

    // Return the number of periods that we can currently count.
    inline size_t size (void) const
    {
        return m_periods.size();
    }

    // Return the number of entries that can be averaged right now.
    inline size_t stored_periods (void) const
    {
        if (m_firstPeriod <= m_nextPeriod)
            return (m_nextPeriod - m_firstPeriod);
        return (m_nextPeriod - m_periods.begin()) + (m_periods.end() - m_firstPeriod);
    }

    // Return the total number of periods counted since creation or the last reset.
    inline size_t total_periods (void) const
    {
        return m_totalPeriods;
    }

    // Return the active period time in milliseconds.
    inline size_t active_period (void) const
    {
        return std::chrono::duration_cast<std::chrono::milliseconds>(Clock::now() - m_lastCountTime).count();
    }

    // Count the period since the last counted period.
    inline size_t count (void)
    {
        const auto now = Clock::now();
        *m_nextPeriod = now - m_lastCountTime;
        m_lastCountTime = now;

        const bool firstIsNext = m_nextPeriod == m_firstPeriod;
        
        if (m_periods.end() == ++m_nextPeriod)
            m_nextPeriod = m_periods.begin();

        if (m_nextPeriod == m_firstPeriod)
        {
            if (m_periods.end() == ++m_firstPeriod)
                m_firstPeriod = m_periods.begin();
        }

        return ++m_totalPeriods;
    }

    // Determine the average period.
    inline typename Clock::duration average_period (void) const
    {
        // Summing variables.
        typename Clock::duration periodSum = Clock::duration::zero();
        auto sumPeriod = [&periodSum](typename Clock::duration p) { periodSum += p; };
        
        size_t periodCount = stored_periods();
        if (m_firstPeriod <= m_nextPeriod)
        {
            periodCount = (m_nextPeriod - m_firstPeriod);
            std::for_each<PeriodConstIter>(m_firstPeriod, m_nextPeriod, sumPeriod);
        }
        else
        {
            periodCount = (m_nextPeriod - m_periods.begin()) + (m_periods.end() - m_firstPeriod);
            std::for_each<PeriodConstIter>(m_periods.begin(), m_nextPeriod, sumPeriod);
            std::for_each(m_firstPeriod, m_periods.end(), sumPeriod);
        }

        // If no periods are set, the average is "infinite".
        if (!periodCount)
            return Clock::duration::max();

        // Simple division.
        return periodSum / periodCount;
    }

    // Determine the average rate in Hz.
    inline double average_rate (void) const
    {
        const double ns_averagePeriod = (double)std::chrono::nanoseconds(average_period()).count();
        const double nsPerSecond = (double)std::chrono::nanoseconds( std::chrono::seconds(1) ).count();
        return nsPerSecond / ns_averagePeriod;
    }
};

#endif // INCLUDED__SIMPLE__TIMERS__HPP
